// Copyright (c) 2010-2019 The Regents of the University of Michigan
// This file is from the freud project, released under the BSD 3-Clause License.

#include <iostream>

#include "Box.h"
#include "Index1D.h"

using namespace std;

/*! \file Box.cc
    \brief Represents simulation boxes and contains helpful wrapping functions.
*/

namespace freud { namespace box {

}; }; // end namespace freud::box
