// Copyright (c) 2010-2019 The Regents of the University of Michigan
// This file is from the freud project, released under the BSD 3-Clause License.

#include "NeighborQuery.h"

namespace freud { namespace locality {

const NeighborPoint NeighborQueryIterator::ITERATOR_TERMINATOR(-1, -1, 0);

}; }; // end namespace freud::locality
